#ifndef _osapi_core_
#define _osapi_core_ 
#include <stdarg.h>
#define OS_SEM_FULL 1
#define OS_SEM_EMPTY 0
#define OS_FP_ENABLED 1
typedef struct
{
    char name [OS_MAX_API_NAME];
    uint32 creator;
    uint32 stack_size;
    uint32 priority;
    uint32 OStask_id;
}OS_task_prop_t;
typedef struct
{
    char name [OS_MAX_API_NAME];
    uint32 creator;
}OS_queue_prop_t;
typedef struct
{
    char name [OS_MAX_API_NAME];
    uint32 creator;
    int32 value;
}OS_bin_sem_prop_t;
typedef struct
{
    char name [OS_MAX_API_NAME];
    uint32 creator;
    int32 value;
}OS_count_sem_prop_t;
typedef struct
{
    char name [OS_MAX_API_NAME];
    uint32 creator;
}OS_mut_sem_prop_t;
typedef struct
{
    uint32 seconds;
    uint32 microsecs;
}OS_time_t;
typedef struct
{
    uint32 free_bytes;
    uint32 free_blocks;
    uint32 largest_free_block;
}OS_heap_prop_t;
#define OS_ERROR_NAME_LENGTH 35
typedef char os_err_name_t[OS_ERROR_NAME_LENGTH];
typedef void osal_task;
typedef osal_task ((*osal_task_entry)(void));
typedef void (*OS_ArgCallback_t)(uint32 object_id, void *arg);
void OS_Application_Startup(void);
int32 OS_API_Init (void);
void OS_IdleLoop (void);
void OS_DeleteAllObjects (void);
void OS_ApplicationShutdown (uint8 flag);
int32 OS_TaskCreate (uint32 *task_id, const char *task_name,
                                osal_task_entry function_pointer,
                                uint32 *stack_pointer,
                                uint32 stack_size,
                                uint32 priority, uint32 flags);
int32 OS_TaskDelete (uint32 task_id);
void OS_TaskExit (void);
void OS_TaskExitAndShutdown (void);
int32 OS_TaskInstallDeleteHandler(osal_task_entry function_pointer);
int32 OS_TaskDelay (uint32 millisecond);
int32 OS_TaskSetPriority (uint32 task_id, uint32 new_priority);
int32 OS_TaskRegister (void);
uint32 OS_TaskGetId (void);
int32 OS_TaskGetIdByName (uint32 *task_id, const char *task_name);
int32 OS_TaskGetInfo (uint32 task_id, OS_task_prop_t *task_prop);
int32 OS_QueueCreate (uint32 *queue_id, const char *queue_name,
                                uint32 queue_depth, uint32 data_size, uint32 flags);
int32 OS_QueueDelete (uint32 queue_id);
int32 OS_QueueGet (uint32 queue_id, void *data, uint32 size,
                                uint32 *size_copied, int32 timeout);
int32 OS_QueuePut (uint32 queue_id, const void *data, uint32 size,
                                uint32 flags);
int32 OS_QueueGetIdByName (uint32 *queue_id, const char *queue_name);
int32 OS_QueueGetInfo (uint32 queue_id, OS_queue_prop_t *queue_prop);
int32 OS_BinSemCreate (uint32 *sem_id, const char *sem_name,
                                uint32 sem_initial_value, uint32 options);
int32 OS_BinSemFlush (uint32 sem_id);
int32 OS_BinSemGive (uint32 sem_id);
int32 OS_BinSemTake (uint32 sem_id);
int32 OS_BinSemTimedWait (uint32 sem_id, uint32 msecs);
int32 OS_BinSemDelete (uint32 sem_id);
int32 OS_BinSemGetIdByName (uint32 *sem_id, const char *sem_name);
int32 OS_BinSemGetInfo (uint32 sem_id, OS_bin_sem_prop_t *bin_prop);
int32 OS_CountSemCreate (uint32 *sem_id, const char *sem_name,
                                uint32 sem_initial_value, uint32 options);
int32 OS_CountSemGive (uint32 sem_id);
int32 OS_CountSemTake (uint32 sem_id);
int32 OS_CountSemTimedWait (uint32 sem_id, uint32 msecs);
int32 OS_CountSemDelete (uint32 sem_id);
int32 OS_CountSemGetIdByName (uint32 *sem_id, const char *sem_name);
int32 OS_CountSemGetInfo (uint32 sem_id, OS_count_sem_prop_t *count_prop);
int32 OS_MutSemCreate (uint32 *sem_id, const char *sem_name, uint32 options);
int32 OS_MutSemGive (uint32 sem_id);
int32 OS_MutSemTake (uint32 sem_id);
int32 OS_MutSemDelete (uint32 sem_id);
int32 OS_MutSemGetIdByName (uint32 *sem_id, const char *sem_name);
int32 OS_MutSemGetInfo (uint32 sem_id, OS_mut_sem_prop_t *mut_prop);
int32 OS_Milli2Ticks (uint32 milli_seconds);
int32 OS_Tick2Micros (void);
int32 OS_GetLocalTime (OS_time_t *time_struct);
int32 OS_SetLocalTime (OS_time_t *time_struct);
int32 OS_ExcAttachHandler (uint32 ExceptionNumber,
                                void (*ExceptionHandler)(uint32, const void *,uint32),
                                int32 parameter);
int32 OS_ExcEnable (int32 ExceptionNumber);
int32 OS_ExcDisable (int32 ExceptionNumber);
int32 OS_FPUExcAttachHandler (uint32 ExceptionNumber, void * ExceptionHandler ,
                                 int32 parameter);
int32 OS_FPUExcEnable (int32 ExceptionNumber);
int32 OS_FPUExcDisable (int32 ExceptionNumber);
int32 OS_FPUExcSetMask (uint32 mask);
int32 OS_FPUExcGetMask (uint32 *mask);
int32 OS_IntAttachHandler (uint32 InterruptNumber, osal_task_entry InterruptHandler, int32 parameter);
int32 OS_IntUnlock (int32 IntLevel);
int32 OS_IntLock (void);
int32 OS_IntEnable (int32 Level);
int32 OS_IntDisable (int32 Level);
int32 OS_IntSetMask (uint32 mask);
int32 OS_IntGetMask (uint32 *mask);
int32 OS_IntAck (int32 InterruptNumber);
int32 OS_ShMemInit (void);
int32 OS_ShMemCreate (uint32 *Id, uint32 NBytes, const char* SegName);
int32 OS_ShMemSemTake (uint32 Id);
int32 OS_ShMemSemGive (uint32 Id);
int32 OS_ShMemAttach (cpuaddr * Address, uint32 Id);
int32 OS_ShMemGetIdByName (uint32 *ShMemId, const char *SegName );
int32 OS_HeapGetInfo (OS_heap_prop_t *heap_prop);
int32 OS_GetErrorName (int32 error_num, os_err_name_t* err_name);
void OS_printf( const char *string, ...) OS_PRINTF(1,2);
void OS_printf_disable(void);
void OS_printf_enable(void);
void OS_ApplicationExit(int32 Status);
#endif
